import pyarrow as pa
import numpy as np
import pyarrow.compute as c

# Cleaning functions
def clean_num(arr, impute=0.0, clip_min=None, clip_max=None):
    return (
        pa.array(
            np.nan_to_num(
                arr.to_numpy(zero_copy_only=False).astype(np.float64), nan=impute
            ).clip(clip_min, clip_max)
        ),
    )


def clean_cat(arr, categories=[]):
    arr = arr.cast(pa.string()).dictionary_encode()
    dic = arr.dictionary.to_pylist()
    if categories:
        d = {
            i: (categories.index(v) + 1 if v in categories else 0)
            for i, v in enumerate(dic)
        }
        d[-1] = 0  # NULLs -> 0
        return (
            pa.array(np.vectorize(d.get)(arr.indices.fill_null(-1).to_numpy())),
            ["Unknown"] + categories,
        )
    else:
        return (
            c.add(arr.indices, pa.array([1], type=pa.int32())[0]).fill_null(0),
            ["Unknown"] + dic,
        )


def clean_hot(arr, categories=[], drop_first=False):
    arr = arr.cast(pa.string())
    if categories:
        clns = [c.equal(arr, v).fill_null(False) for v in categories]
    else:
        categories = [u for u in arr.unique().to_pylist() if u]
        clns = [c.equal(arr, v).fill_null(False) for v in categories]
    return clns[(1 if drop_first else 0) :], categories[(1 if drop_first else 0) :]


# Cleaning Classes
class NumericalColumn:
    def __init__(
        self, name, impute="mean", clip=True, v_min=None, v_mean=None, v_max=None
    ):
        self.name, self.impute, self.clip = name, impute, clip
        self.measured = any([v_min, v_mean, v_max])
        self.mean, self.min, self.max = (v_mean or 0), (v_min or 0), (v_max or 0)

    def to_dict(self):
        return {
            "name": self.name,
            "type": "numerical",
            "impute": self.impute,
            "clip": self.clip,
            "v_min": self.min,
            "v_mean": self.mean,
            "v_max": self.max,
        }

    def update(self, arr):
        self.mean = float(c.mean(arr).as_py())
        minmax = c.min_max(arr)
        self.min, self.max = float(minmax["min"].as_py()), float(minmax["max"].as_py())

    def value(self):
        if self.impute == "mean":
            return self.mean
        elif self.impute == "min":
            return self.min
        elif self.impute == "max":
            return self.max
        else:
            raise Exception("{} is not a valid impute method".format(self.impute))

    def clean(self, arr):
        if not self.measured:
            self.update(arr)
        (cln,) = clean_num(
            arr,
            impute=self.value(),
            clip_min=(self.min if self.clip else None),
            clip_max=(self.max if self.clip else None),
        )
        return cln, None


class CategoricalColumn:
    def __init__(self, name, method, categories=[]):
        self.name, self.method, self.categories = name, method, categories
        self.measured = True if categories else False

    def to_dict(self):
        return {
            "name": self.name,
            "type": "categorical",
            "method": self.method,
            "categories": self.categories,
        }

    def update(self, categories):
        self.categories = self.categories + [
            c for c in categories if c not in self.categories
        ]

    def clean(self, arr):
        if self.method == "one_hot":
            cln, cats = clean_hot(arr, categories=self.categories)
        else:
            cln, cats = clean_cat(arr, categories=self.categories)
        if not self.measured:
            self.categories = cats
        return cln, cats


class TableCleaner:
    def __init__(self):
        self.columns = []

    def to_dict(self):
        return [column.to_dict() for column in self.columns]

    def from_dict(self, columns):
        for column in columns:
            t = column.pop("type")
            if t == "numerical":
                self.columns.append(NumericalColumn(**column))
            else:
                self.columns.append(CategoricalColumn(**column))
        return self

    def register_numeric(self, name, impute="mean", clip=True):
        self.columns.append(NumericalColumn(name, impute, clip))

    def register_label(self, name, categories=[]):
        self.columns.append(
            CategoricalColumn(name, method="label", categories=categories)
        )

    def register_one_hot(self, name, categories=[]):
        self.columns.append(
            CategoricalColumn(name, method="one_hot", categories=categories)
        )

    def clean_column(self, table, column):
        arr = table.column(column.name).combine_chunks()
        cln, cats = column.clean(arr)
        if column.__dict__.get("method", "") == "one_hot":
            return [column.name + "_" + cat for cat in cats], cln
        else:
            return [column.name], [cln]

    def clean_table(self, table, label=None):
        keys, arrays = [], []
        for column in self.columns:
            k, a = self.clean_column(table, column)
            keys.extend(k)
            arrays.extend(a)
        if label:
            return pa.Table.from_arrays(arrays, names=keys), table.column(label)
        else:
            return pa.Table.from_arrays(arrays, names=keys)

    def split(self, X, y=None, test_size=0.2):
        mask = np.random.rand(X.num_rows) > test_size
        while np.all(mask):  # [True, True, True] is invalid
            mask = np.random.rand(X.num_rows) > test_size
        idxs, not_idxs = np.where(mask)[0], np.where(~mask)[0]
        return (
            X.take(idxs),
            X.take(not_idxs),
            y.take(idxs),
            y.take(not_idxs),
        )  # X_train, X_test, y_train, y_test



def is_int(value):
    try:
        int_value = int(value)
        return True
    except ValueError:
        return False


def is_float(value):
    try:
        int_value = float(value)
        return True
    except ValueError:
        return False



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implement a Relation, this analogous to a database table. It is called a Relation
in-line with terminology for Relational Algreba.

Internally the data for the Relation is stored as a list of Tuples and a dictionary
records the schema and other key information about the Relation. Tuples are faster
than Dicts for accessing data.

Some naive benchmarking with 135k records (with 269 columns) vs lists of dictionaries:
- Relation is 30% smaller (this doesn't hold true for all datasets)
- Relation was deduplicated 90% faster
- Relation selection was 75% faster


- self.data is a list of tuples
- self.header is a dictionary
    {
        "attribute": {
            "type": type (domain)
            "min": smallest value in this attribute
            "max": largest value in this attribute
            "count": count of non-null values in attribute
            "unique": number of unique values in this attribute
            "nullable": are there nulls
        }
    }

    Only "type" can be assumed to be present, especially after a select operation
"""
import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], "../.."))

from typing import Iterable, Tuple
from opteryx.engine.attribute_types import (
    OPTERYX_TYPES,
    PARQUET_TYPES,
    PYTHON_TYPES,
    coerce_types,
)
from opteryx.exceptions import MissingDependencyError


class Relation:

    __slots__ = ("header", "data", "name")

    def __init__(
        self, data: Iterable[Tuple] = [], *, header: dict = {}, name: str = None
    ):
        """
        Create a Relation.

        Parameters:
            data: Iterable
                An iterable which is the data in the Relation
            header: Dictionary (optional)
                Schema and profile information for this Relation
            name: String (optional)
                A handle for this Relation, cosmetic only
        """
        self.data = data
        self.header = header
        self.name = name

    def apply_selection(self, predicate):
        """
        Apply a Selection operation to a Relation, this filters the data in the
        Relation to just the entries which match the predicate.

        Parameters:
            predicate (list or tuple):
                A DNF structured filter predicate.

        Returns:
            Relation
        """
        # selection invalidates what we thought we knew about counts etc
        new_header = {k: {"type": v.get("type")} for k, v in self.header.items()}
        return Relation(filter(predicate, self.data), header=new_header)

    def apply_projection(self, attributes):
        if not isinstance(attributes, (list, tuple)):
            attributes = [attributes]
        attribute_indices = []
        new_header = {k: self.header.get(k) for k in attributes}
        for index, attribute in enumerate(self.header.keys()):
            if attribute in attributes:
                attribute_indices.append(index)

        def _inner_projection():
            for tup in self.data:
                yield tuple([tup[indice] for indice in attribute_indices])

        return Relation(_inner_projection(), header=new_header)

    def __getitem__(self, attributes):
        """
        Select the attributes from the Relation, alias for .apply_projection called
        like this Relation["column"]
        """
        return self.apply_projection(attributes)

    def materialize(self):
        if not isinstance(self.data, list):
            self.data = list(self.data)

    def count(self):
        self.materialize()
        return len(self.data)

    @property
    def shape(self):
        self.materialize()
        return (len(self.header), self.count())

    def distinct(self):
        """
        Return a new Relation with only unique values
        """
        hash_list = {}
        self.materialize()

        def do_dedupe(data):
            for item in data:
                hashed_item = hash(item)
                if hashed_item not in hash_list:
                    yield item
                    hash_list[hashed_item] = True

        return Relation(do_dedupe(self.data), header=self.header)

    def from_dictionaries(self, dictionaries, header=None):

        from operator import itemgetter

        def types(dictionaries):
            response = {}
            for row in dictionaries:
                for k, v in row.items():
                    value_type = PYTHON_TYPES.get(str(type(v)), OPTERYX_TYPES.OTHER)
                    if k not in response:
                        response[k] = {"type": value_type}
                    elif response[k] != {"type": value_type}:
                        response[k] = {"type": OPTERYX_TYPES.OTHER}
            return response

        if not header:
            self.header = types(dictionaries)
        self.data = [
            tuple([coerce_types(row.get(k)) for k in self.header.keys()])
            for row in dictionaries
        ]

    def attributes(self):
        return [k for k in self.header.keys()]

    def rename_attribute(self, current_name, new_name):
        self.header[new_name] = self.header.pop(current_name)

    def __str__(self):
        return f"{self.name or 'Relation'} ({', '.join([k + ':' + v.get('type') for k,v in self.header.items()])})"

    def __len__(self):
        """
        Alias for .count
        """
        return self.count()

    def serialize(self):
        """
        Serialize to Parquet, this is used for communicating and saving the Relation.
        """
        try:
            import pyarrow.parquet as pq
            import pyarrow.json
        except ImportError:  # pragma: no cover
            raise MissingDependencyError(
                "`pyarrow` is missing, please install or include in requirements.txt"
            )

        import io

        # load into pyarrow as a JSON dataset
        in_pyarrow_buffer = pyarrow.json.read_json(io.BytesIO(self.to_json()))

        # then we convert the pyarrow table to parquet
        pq_temp_buffer = io.BytesIO()
        pq.write_table(in_pyarrow_buffer, pq_temp_buffer, compression="ZSTD")

        # the read out the parquet formatted data
        pq_temp_buffer.seek(0, 0)
        buffer = pq_temp_buffer.read()
        pq_temp_buffer.close()

        return buffer

    def to_json(self):
        """
        Convert the relation to a JSONL byte string
        """
        import orjson

        return b"\n".join(map(orjson.dumps, self.fetchall()))

    def fetchone(self, offset: int = 0):
        self.materialize()
        try:
            return dict(zip(self.header.keys(), self.data[offset]))
        except IndexError:
            return None

    def fetchmany(self, size: int = 100, offset: int = 0):
        keys = self.header.keys()
        self.materialize()

        def _inner_fetch():
            for index in range(offset, min(len(self.data), offset + size)):
                yield dict(zip(keys, self.data[index]))

        return list(_inner_fetch())

    def fetchall(self, offset: int = 0):
        keys = self.header.keys()
        self.materialize()

        def _inner_fetch():
            for index in range(offset, len(self.data)):
                yield dict(zip(keys, self.data[index]))

        return list(_inner_fetch())

    def collect_column(self, column):
        def get_column(column):
            for index, attribute in enumerate(self.header.keys()):
                if attribute == column:
                    return index
            raise Exception("Column not found")

        def _inner_fetch(column):
            for index in range(len(self.data)):
                yield self.data[index][column]

        return list(_inner_fetch(get_column(column)))

    @staticmethod
    def deserialize(stream):
        """
        Deserialize from a stream of bytes which is a Parquet file.

        Return a Relation with a basic schema.
        """
        try:
            import pyarrow.parquet as pq  # type:ignore
        except ImportError:  # pragma: no cover
            raise MissingDependencyError(
                "`pyarrow` is missing, please install or include in requirements.txt"
            )
        if not hasattr(stream, "read") or isinstance(stream, bytes):
            import io

            stream = io.BytesIO(stream)
        table = pq.read_table(stream)

        def _inner(table):
            for batch in table.to_batches():
                dict_batch = batch.to_pydict()
                for index in range(len(batch)):
                    yield tuple([v[index] for k, v in dict_batch.items()])

        def pq_schema_to_rel_schema(pq_schema):
            schema = {}
            for attribute in pq_schema:
                schema[attribute.name] = {
                    "type": PARQUET_TYPES.get(
                        str(attribute.type), OPTERYX_TYPES.OTHER
                    ).value
                }
            return schema

        return Relation(_inner(table), header=pq_schema_to_rel_schema(table.schema))

    @staticmethod
    def load(file):
        with open(file, "rb") as pq_file:
            return Relation.deserialize(pq_file)


if __name__ == "__main__":

    from opteryx.utils.timer import Timer

    with Timer("load"):
        r = Relation.load("tests/data/parquet/tweets.parquet")
    r.materialize()

    r.data = r.data * 10

    print(r.header)
    print(r.data[10])

    print(r.count())
    print(r.attributes())
    r.distinct()
    print(r.attributes())
    print(r.apply_projection(["user_verified"]).distinct().count())
    print(len(r.fetchall()))

    with Timer("s"):
        s = r.serialize()

    with Timer("d"):
        s = Relation.deserialize(s)

    print(s.count())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Decode files from a raw binary format to a PyArrow Table.
"""


def zstd_decoder(stream, projection):
    """
    Read zstandard compressed JSONL files
    """
    import zstandard

    with zstandard.open(stream, "rb") as file:
        return jsonl_decoder(file, projection)


def parquet_decoder(stream, projection):
    """
    Read parquet formatted files
    """
    import pyarrow.parquet as pq

    table = pq.read_table(stream, columns=projection)
    return table


def orc_decoder(stream, projection):
    """
    Read orc formatted files
    """
    import pyarrow.orc as orc

    orc_file = orc.ORCFile(stream)
    table = orc_file.read(columns=projection)
    return table


def jsonl_decoder(stream, projection):

    import pyarrow.json

    table = pyarrow.json.read_json(stream)

    # the read doesn't support projection, so do it now
    if projection:
        table = table.select(projection)

    return table


def arrow_decoder(stream, projection):

    import pyarrow.feather as pf

    table = pf.read_table(stream, columns=projection)
    return table



"""
Selection Node

This is a SQL Query Execution Plan Node.

This Node eliminates elminiates records which do not match a predicate using a
DNF (Disjunctive Normal Form) interpretter. 

Predicates in the same list are joined with an AND/Intersection (all must be True)
and predicates in adjacent lists are joined with an OR/Union (any can be True).
This allows for non-trivial filters to be written with just tuples and lists.

The predicates are in _tuples_ in the form (`key`, `op`, `value`) where the `key`
is the value looked up from the record, the `op` is the operator and the `value`
is a literal.
"""
import numpy
from typing import Union, Iterable
from pyarrow import Table
from opteryx.engine.query_statistics import QueryStatistics
from opteryx.engine.planner.operations.base_plan_node import BasePlanNode
from opteryx.third_party.pyarrow_ops import ifilters


class InvalidSyntaxError(Exception):
    pass


def _evaluate(predicate: Union[tuple, list], table: Table) -> bool:

    # If we have a tuple extract out the key, operator and value and do the evaluation
    if isinstance(predicate, tuple):
        # filters from pyarrow_ops only filters on a single predicate
        return ifilters(table, predicate)

    # If we have a list, we're going to recurse and call ourselves with the items in
    # the list
    if isinstance(predicate, list):
        # Are all of the entries tuples?
        # We AND them together
        mask = None
        if all([isinstance(p, tuple) for p in predicate]):
            for p in predicate:
                if mask is None:
                    # The first time round we either set the mask to the first set of
                    # values, or we initialize the mask beforehand to all True.
                    mask = _evaluate(p, table)
                else:
                    mask = numpy.intersect1d(mask, _evaluate(p, table))
            return mask

        # Are all of the entries lists?
        # We OR them together
        if all([isinstance(p, list) for p in predicate]):
            for p in predicate:
                if mask is None:
                    # The first time round we either set the mask to the first set of
                    # values, or we initialize the mask beforehand to all empty.
                    mask = _evaluate(p, table)
                else:
                    mask = numpy.union1d(mask, _evaluate(p, table))
            return mask

        # if we're here the structure of the filter is wrong
        raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover

    raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover


class SelectionNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        self._filter = config.get("filter")

    def __repr__(self):
        return str(self._filter)

    def execute(self, data_pages: Iterable) -> Iterable:

        if self._filter is None:
            yield from data_pages

        else:
            for page in data_pages:
                mask = _evaluate(self._filter, page)
                yield page.take(mask)



"""
Grouping Node

This is a SQL Query Execution Plan Node.

This performs aggregations, both of grouped and non-grouped data. 

This is a greedy operator - it consumes all the data before responding.

This algorithm is a balance of performance, it is much slower than a groupby based on
the pyarrow_ops library for datasets with a high number of duplicate values (e.g.
grouping by a boolean column) - on a 10m record set, timings are 10:1 (removing raw
read time - e.g. 30s:21s where 20s is the read time).

But, on high cardinality data (nearly unique columns), the performance is much faster,
on a 10m record set, timings are 1:400 (50s:1220s where 20s is the read time).
"""
import io
import orjson
import pyarrow.json
import numpy as np
from typing import Iterable
from opteryx.engine import QueryStatistics
from opteryx.engine.planner.operations import BasePlanNode

# these functions can be applied to each group
INCREMENTAL_AGGREGATES = {
    "MIN": lambda x, y: min(x, y),
    "MAX": lambda x, y: max(x, y),
    "SUM": lambda x, y: x + y,
}


# these functions need the whole dataset
WHOLE_AGGREGATES = {
    "AVG": np.mean,
    "MEDIAN": np.median,
    "PRODUCT": np.prod,
    "STDDEV_POP": np.std,
    "VAR_POP": np.var,
    "FIRST": lambda a: a[0],
    "LAST": lambda a: a[-1],
    # range - difference between min and max
    # percent - each group has the relative portion calculated
    # list - return a list of the items in the list
    # quantile
    # approx_distinct
    # approx_quantile
}


def _incremental(x, y, function):
    if function in INCREMENTAL_AGGREGATES:
        return INCREMENTAL_AGGREGATES[function](x, y)
    return np.concatenate(x, y)



JSON_TYPES = {
    np.bool_: bool,
    np.int64: int,
    np.float64: float
}


def _serializer(obj):
    return JSON_TYPES[type(obj)](obj)


def _map(table, collect_columns):
    # if we're aggregating on * we don't care about the data, just the number of
    # records
    if collect_columns == ["*"]:
        for i in range(table.num_rows):
            yield (("*", "*"),)
        return

    if "*" in collect_columns:
        collect_columns = [c for c in collect_columns if c != "*"]
    arr = [c.to_numpy() for c in table.select(collect_columns)]

    for row_index in range(len(arr[0])):
        ret = []
        for column_index, column_name in enumerate(collect_columns):
            ret.append((column_name, arr[column_index][row_index]))
        yield ret


class AggregateNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):

        from opteryx.engine.attribute_types import TOKEN_TYPES

        self._aggregates = []
        self._groups = config.get("groups", [])
        self._project = self._groups.copy()
        aggregates = config.get("aggregates", [])
        for attribute in aggregates:
            if "aggregate" in attribute:
                self._aggregates.append(attribute)
                argument = attribute["args"][0]
                column = argument[0]
                if argument[1] == TOKEN_TYPES.WILDCARD:
                    column = "*"
                self._project.append(column)
            else:
                self._project.append(attribute)

        self._project = list(set(self._project))

    def __repr__(self):
        return str(self._aggregates)

    def greedy(self):
        return True

    def execute(self, groups: Iterable) -> Iterable:

        from collections import defaultdict

        collector = defaultdict(dict)

        for page in groups:

            for group in _map(page, self._project):

                for aggregrator in self._aggregates:

                    attribute = aggregrator["args"][0][0]
                    function = aggregrator["aggregate"]
                    column_name = f"{function}({attribute})"
                    value = [v for k,v in group if k == attribute]
                    if len(value) == 1:
                        value = value[0]

                    # this is needed for situations where the select column isn't selecting
                    # the columns in the group by
                    collection = tuple(
                        [
                            (
                                k,
                                v,
                            )
                            for k, v in group
                            if k in self._groups
                        ]
                    )

                    group_collector = collector[collection]

                    # Add the responses to the collector
                    # if it's COUNT(*)
                    if column_name == "COUNT(Wildcard)":
                        if "COUNT(*)" in group_collector:
                            group_collector["COUNT(*)"] += 1
                        else:
                            group_collector["COUNT(*)"] = 1
                    elif function == "COUNT":
                        if value:
                            if column_name in group_collector:
                                group_collector[column_name] += 1
                            else:
                                group_collector[column_name] = 1
                    # if we have some information collected for this group,
                    # incremental update
                    elif column_name in group_collector:
                        group_collector[column_name] = _incremental(
                            value,
                            group_collector[column_name],
                            function,
                        )
                    # otherwise, it's new, so seed the collection with the initial value
                    else:
                        group_collector[column_name] = value

                    collector[collection] = group_collector
                # TODO: if we're going to cap the number of groups we collect, do it here

        # TODO: do any whole aggregate functions

        import time

        from pyarrow.json import ReadOptions

        ro = ReadOptions(block_size=3 * 1024 * 1024)

        buffer = bytearray()
        t = time.time_ns()
        for collected, record in collector.items():
            # we can't load huge json docs into pyarrow, so we chunk it
            if len(buffer) > (2 * 1024 * 1024):  # 4Mb
                table = pyarrow.json.read_json(io.BytesIO(buffer), read_options=ro)
                yield table
                buffer = bytearray()
            for k, v in collected:
                if hasattr(v, "as_py"):
                    v = v.as_py()
                record[k] = v
            buffer.extend(orjson.dumps(record, default=_serializer))

        if len(buffer) > 0:
            table = pyarrow.json.read_json(io.BytesIO(buffer))
            yield table

        # timing over a yield is pointless
        # print("building group table", (time.time_ns() - t) / 1e9)



from opteryx.storage import BasePartitionScheme
from typing import Union, Tuple


class DefaultPartitionScheme(BasePartitionScheme):
    def __init__(self, format: Union[Tuple, str]):
        if not isinstance(format, (list, set, tuple)):
            self._format = [format]
        else:
            self._format = format

    def partition_format(self):
        return "/".join(self._format)

    def filter_blobs(self, list_of_blobs):
        return list_of_blobs



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyarrow import compute
import datetime
import numpy

 

FUNCTIONS = {

    # VECTORIZED FUNCTIONS
    "LENGTH": compute.utf8_length, # LENGTH(str) -> int
    "UPPER": compute.utf8_upper, # UPPER(str) -> str
    "LOWER": compute.utf8_lower, # LOWER(str) -> str
    "TRIM": compute.utf8_trim_whitespace, # TRIM(str) -> str

}

if __name__ == "__main__":

    import sys
    import os

    sys.path.insert(1, os.path.join(sys.path[0], "../../.."))

    from opteryx.samples import planets
    from opteryx.third_party.pyarrow_ops import head
    from opteryx.utils import dates

    from pyarrow import compute

    p = planets()

    #print(p["name"])

    #print(FUNCTIONS["UPPER"](p["name"]))
    #print(list(FUNCTIONS["LEFT"](p["name"], 3)))
    #print(list(FUNCTIONS["LEFT"](FUNCTIONS["RIGHT"](p["name"], 4), 2)))

    print(compute.year(numpy.datetime64('2021-01-01').astype(datetime.datetime)))


import random

"""
Functions for generating the cyclic group [0,...n-1]. Use instead of
random.shuffle() or similar.

Functions:
    pyudorandom(n) <- generate the numbers in 0,...n-1
    bin_gcd(a, b) <- calculate the gcd of a and b fast

"""


def items(ls):
    """
    Yields the elements of ls in a pseudorandom fashion.

    """
    num = len(ls)
    if num == 0:
        return
    for i in indices(num):
        yield ls[i]


def shuffle(ls):
    """
    Takes a list ls and returns a new list with the elements of ls
    in a new order.

    """
    return list(items(ls))


def indices(n):
    """
    Generates the cyclic group 0 through n-1 using a number
    which is relative prime to n.

    """
    rand = find_gcd_one(n)
    i = 1
    while i <= n:
        yield i * rand % n
        i += 1


def find_gcd_one(n):
    """
    Find a number between 1 and n that has gcd with n equal 1.

    """
    while True:
        rand = int(random.random() * n)
        if bin_gcd(rand, n) == 1:
            return rand


def bin_gcd(a, b):
    """
    Return the greatest common divisor of a and b using the binary
    gcd algorithm.

    """
    if a == b or b == 0:
        return a
    if a == 0:
        return b

    if not a & 1:
        if not b & 1:
            return bin_gcd(a >> 1, b >> 1) << 1
        else:
            return bin_gcd(a >> 1, b)
    if not b & 1:
        return bin_gcd(a, b >> 1)
    if a > b:
        return bin_gcd((a - b) >> 1, b)

    return bin_gcd((b - a) >> 1, a)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

"""
import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], "../../.."))


def canonical_sql(ast, indent=False):

    statement = ""

    # SELECT
    statement += ast[0]["Query"]["body"]["Select"]["projection"]
    # FROM
    # WHERE
    # GROUP BY
    # HAVING
    # ORDER BY
    # LIMIT

    return statement



from functools import lru_cache
import re
import datetime
from typing import Optional, Union

TIMEDELTA_REGEX = (
    r"((?P<days>-?\d+)d)?"
    r"((?P<hours>-?\d+)h)?"
    r"((?P<minutes>-?\d+)m)?"
    r"((?P<seconds>-?\d+)s)?"
)
TIMEDELTA_PATTERN = re.compile(TIMEDELTA_REGEX, re.IGNORECASE)


def extract_date(value):
    if isinstance(value, str):
        value = parse_iso(value)
    if isinstance(value, (datetime.date, datetime.datetime)):
        return datetime.date(value.year, value.month, value.day)
    return datetime.date.today()


# based on:
# https://gist.github.com/santiagobasulto/698f0ff660968200f873a2f9d1c4113c#file-parse_timedeltas-py
def parse_delta(delta: str) -> datetime.timedelta:
    """
    Parses a human readable timedelta (3d5h19m) into a datetime.timedelta.

    Delta includes:
    * Xd days
    * Xh hours
    * Xm minutes
    * Xs seconds

    Values can be negative following timedelta's rules. Eg: -5h-30m
    """
    match = TIMEDELTA_PATTERN.match(delta)
    if match:
        parts = {k: int(v) for k, v in match.groupdict().items() if v}
        return datetime.timedelta(**parts)
    return datetime.timedelta(seconds=0)


@lru_cache(128)
def parse_iso(value):
    DATE_SEPARATORS = {"-", ":"}
    # date validation at speed is hard, dateutil is great but really slow, this is fast
    # but error-prone. It assumes it is a date or it really nothing like a date.
    # Making that assumption - and accepting the consequences - we can convert upto
    # three times faster than dateutil.
    #
    # valid formats (not exhaustive):
    #
    #   YYYY-MM-DD                 <- date
    #   YYYY-MM-DD HH:MM           <- date and time, no seconds
    #   YYYY-MM-DDTHH:MM           <- date and time, T separator
    #   YYYY-MM-DD HH:MM:SS        <- date and time with seconds
    #   YYYY-MM-DD HH:MM:SS.mmmm   <- date and time with milliseconds
    #
    # If the last character is a Z, we ignore it.
    try:
        if value[-1] == "Z":
            value = value[:-1]
        val_len = len(value)
        if isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
            return value
        if isinstance(value, str) and 10 <= val_len <= 24:
            if not value[4] in DATE_SEPARATORS or not value[7] in DATE_SEPARATORS:
                return None
            if val_len == 10:
                # YYYY-MM-DD
                return datetime.datetime(
                    *map(int, [value[:4], value[5:7], value[8:10]])
                )
            if val_len >= 16:
                if not (value[10] in ("T", " ") and value[13] in DATE_SEPARATORS):
                    return False
                if val_len == 24 and value[16] in DATE_SEPARATORS and value[19] == ".":
                    # YYYY-MM-DD HH:MM:SS.ssss
                    return datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],  # YYYY
                                value[5:7],  # MM
                                value[8:10],  # DD
                                value[11:13],  # HH
                                value[14:16],  # MM
                                value[17:19],  # SS
                                value[20:24],  # ssss
                            ],
                        )
                    )
                if val_len == 19 and value[16] in DATE_SEPARATORS:
                    # YYYY-MM-DD HH:MM:SS
                    return datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],  # YYYY
                                value[5:7],  # MM
                                value[8:10],  # DD
                                value[11:13],  # HH
                                value[14:16],  # MM
                                value[17:19],  # SS
                            ],
                        )
                    )
                elif val_len == 16:
                    # YYYY-MM-DD HH:MM
                    return datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],
                                value[5:7],
                                value[8:10],
                                value[11:13],
                                value[14:16],
                            ],
                        )
                    )
        return None
    except (ValueError, TypeError):
        return None


def date_range(
    start_date: Optional[Union[str, datetime.date]],
    end_date: Optional[Union[str, datetime.date]],
):
    """
    An interator over a range of dates
    """
    # if dates aren't provided, use today
    end_date = extract_date(end_date)
    start_date = extract_date(start_date)

    if end_date < start_date:  # type:ignore
        raise ValueError(
            "date_range: end_date must be the same or later than the start_date "
        )

    for n in range(int((end_date - start_date).days) + 1):  # type:ignore
        yield start_date + datetime.timedelta(n)  # type:ignore



"""
Projection Node

This is a SQL Query Execution Plan Node.

This Node eliminates columns that are not needed in a Relation. This is also the Node
that performs column renames.
"""
from typing import Iterable
from opteryx.engine.attribute_types import TOKEN_TYPES
from opteryx.engine.query_statistics import QueryStatistics
from opteryx.engine.planner.operations.base_plan_node import BasePlanNode


def replace_wildcards(arg):
    if arg[1] == TOKEN_TYPES.WILDCARD:
        return "*"
    return str(arg[0])


class ProjectionNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        """
        Attribute Projection, remove unwanted columns and performs column renames.
        """
        self._projection = []

        projection = config.get("projection", {"*": "*"})
        # print("projection:", projection)
        for attribute in projection:
            if "aggregate" in attribute:
                self._projection.append(
                    f"{attribute['aggregate']}({','.join([replace_wildcards(a) for a in attribute['args']])})"
                )
            elif "function" in attribute:
                self._projection.append(
                    f"{attribute['function']}({','.join([replace_wildcards(a) for a in attribute['args']])})"
                )
            else:
                self._projection.append(attribute)

    def __repr__(self):
        return str(self._projection)

    def execute(self, data_pages: Iterable) -> Iterable:

        from opteryx.third_party.pyarrow_ops.group import Grouping

        # if we have nothing to do, move along
        if self._projection == ["*"]:
            # print(f"projector yielding *")
            yield from data_pages

        for page in data_pages:

            # allow simple projections using just the list of attributes
            if isinstance(self._projection, (list, tuple, set)):
                # print(f"projector yielding {page.shape}")
                try:
                    yield page.select(list(self._projection))
                except KeyError as e:
                    # print(f"Available columns: {page.column_names}")
                    raise e

            else:
                # we elminimate attributes we don't want
                page = page.select(list(self._projection.keys()))

                # then we rename the attributes
                if any([k != v for k, v in self._projection.items()]):
                    names = [
                        self._projection[a]
                        for a in page.column_names
                        if a in self._projection
                    ]
                    page = page.rename_columns(names)

                # print(f"projector yielding {page.shape}")
                yield page



import pyarrow
from typing import Iterable, List


def fetchmany(pages: Iterable, size: int = 5) -> List[dict]:
    """
    This is the fastest way I've found to do this, on my computer it's about
    14,000 rows per second - fast enough for my use case.
    """

    def _inner_row_reader():
        for page in pages:
            for index in range(page.num_rows):
                row = page.take([index]).to_pydict()
                for k, v in row.items():
                    row[k] = v[0]
                yield row

    index = -1
    for index, row in enumerate(_inner_row_reader()):
        if index == size:
            return
        yield row

    if index < 0:
        yield {}


def fetchone(pages: Iterable) -> dict:
    return fetchmany(pages=pages, limit=1).pop()


def fetchall(pages) -> List[dict]:
    return fetchmany(pages=pages, size=-1)


class Grouping:
    def __init__(self, table, columns):
        # This is a reduced version of the groupby function in
        # https://github.com/TomScheffers/pyarrow_ops
        from opteryx.third_party.pyarrow_ops.helpers import (
            columns_to_array,
            groupify_array,
        )

        self.table = table

        # Initialize array + groupify
        self.arr = columns_to_array(table, columns)
        # dic - the groups
        # counts - the number of items for each group
        # sort indexes - the indexes which would sort this table
        # begin indexes - the first entry for each group
        self.dic, self.counts, self.sort_idxs, self.bgn_idxs = groupify_array(self.arr)

    def __iter__(self):
        for i in range(len(self.dic)):
            idxs = self.sort_idxs[self.bgn_idxs[i] : self.bgn_idxs[i] + self.counts[i]]
            yield self.table.take(idxs)


def groupby(table, by):
    return Grouping(table, by)



#!/usr/bin/env python
# coding:utf-8
# Author:  Mozman
# Purpose: abstract base class for all binary trees
# Created: 03.05.2010
# Copyright (c) 2010-2013 by Manfred Moitzi
# License: MIT License

# Copyright (c) 2012, Manfred Moitzi
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys

PYPY = hasattr(sys, "pypy_version_info")

from .treeslice import TreeSlice
from operator import attrgetter


class _ABCTree(object):
    """
    Abstract-Base-Class for ABCTree and Cython trees.

    The _ABCTree Class
    ==================

    T has to implement following properties
    ---------------------------------------

    count -- get node count

    T has to implement following methods
    ------------------------------------

    get_value(...)
        get_value(key) -> returns value for key

    insert(...)
        insert(key, value) <==> T[key] = value, insert key into T

    remove(...)
        remove(key) <==> del T[key], remove key from T

    clear(...)
        T.clear() -> None.  Remove all items from T.

    iter_items(...)
        iter_items(s, e, [reverse]) -> iterate over all items with keys in range s <= key < e, yielding (k, v) tuple

    foreach(...)
        foreach(f, [order]) -> visit all nodes of tree and call f(k, v) for each node

    pop_item(...)
        T.pop_item() -> (k, v), remove and return some (key, value)

    min_item(...)
        min_item() -> get smallest (key, value) pair of T

    max_item(...)
        max_item() -> get largest (key, value) pair of T

    prev_item(...)
        prev_item(key) -> get (k, v) pair, where k is predecessor to key

    succ_item(...)
        succ_item(key) -> get (k,v) pair as a 2-tuple, where k is successor to key

    floor_item(...)
        floor_item(key) -> get (k, v) pair, where k is the greatest key less than or equal to key

    ceiling_item(...)
        ceiling_item(key) -> get (k, v) pair, where k is the smallest key greater than or equal to key

    Methods defined here
    --------------------

    * __contains__(k) -> True if T has a key k, else False, O(log(n))
    * __delitem__(y) <==> del T[y], del T[s:e], O(log(n))
    * __getitem__(y) <==> T[y], T[s:e], O(log(n))
    * __iter__() <==> iter(T)
    * __len__() <==> len(T), O(1)
    * __max__() <==> max(T), get max item (k,v) of T, O(log(n))
    * __min__() <==> min(T), get min item (k,v) of T, O(log(n))
    * __and__(other) <==> T & other, intersection
    * __or__(other) <==> T | other, union
    * __sub__(other) <==> T - other, difference
    * __xor__(other) <==> T ^ other, symmetric_difference
    * __repr__() <==> repr(T)
    * __setitem__(k, v) <==> T[k] = v, O(log(n))
    * clear() -> None, remove all items from T, , O(n)
    * remove_items(keys) -> None, remove items by keys
    * copy() -> a shallow copy of T, O(n*log(n))
    * discard(k) -> None, remove k from T, if k is present, O(log(n))
    * get(k[,d]) -> T[k] if k in T, else d, O(log(n))
    * is_empty() -> True if len(T) == 0, O(1)
    * keys([reverse]) -> generator for keys of T, O(n)
    * values([reverse]) -> generator for values of  T, O(n)
    * pop(k[,d]) -> v, remove specified key and return the corresponding value, O(log(n))
    * set_default(k[,d]) -> value, T.get(k, d), also set T[k]=d if k not in T, O(log(n))
    * update(E) -> None.  Update T from dict/iterable E, O(E*log(n))

    slicing by keys

    * key_slice(s, e[, reverse]) -> generator for keys of T for s <= key < e, O(n)
    * value_slice(s, e[, reverse]) -> generator for values of T for s <= key < e, O(n)
    * item_slice(s, e[, reverse]) -> generator for items of T for s <= key < e, O(n)
    * T[s:e] -> TreeSlice object, with keys in range s <= key < e, O(n)
    * del T[s:e] -> remove items by key slicing, for s <= key < e, O(n)

    if 's' is None or T[:e] TreeSlice/iterator starts with value of min_key()
    if 'e' is None or T[s:] TreeSlice/iterator ends with value of max_key()
    T[:] is a TreeSlice which represents the whole tree.

    TreeSlice is a tree wrapper with range check and contains no references
    to objects, deleting objects in the associated tree also deletes the object
    in the TreeSlice.

    * TreeSlice[k] -> get value for key k, raises KeyError if k not exists in range s:e
    * TreeSlice[s1:e1] -> TreeSlice object, with keys in range s1 <= key < e1

      * new lower bound is max(s, s1)
      * new upper bound is min(e, e1)

    TreeSlice methods:

    * items() -> generator for (k, v) items of T, O(n)
    * keys() -> generator for keys of T, O(n)
    * values() -> generator for values of  T, O(n)
    * __iter__ <==> keys()
    * __repr__ <==> repr(T)
    * __contains__(key)-> True if TreeSlice has a key k, else False, O(log(n))

    prev/succ operations

    * prev_key(key) -> k, get the predecessor of key, O(log(n))
    * succ_key(key) -> k, get the successor of key, O(log(n))
    * floor_key(key) -> k, get the greatest key less than or equal to key, O(log(n))
    * ceiling_key(key) -> k, get the smallest key greater than or equal to key, O(log(n))

    Heap methods

    * max_key() -> get largest key of T, O(log(n))
    * min_key() -> get smallest key of T, O(log(n))
    * pop_min() -> (k, v), remove item with minimum key, O(log(n))
    * pop_max() -> (k, v), remove item with maximum key, O(log(n))
    * nlargest(i[,pop]) -> get list of i largest items (k, v), O(i*log(n))
    * nsmallest(i[,pop]) -> get list of i smallest items (k, v), O(i*log(n))

    Set methods (using frozenset)

    * intersection(t1, t2, ...) -> Tree with keys *common* to all trees
    * union(t1, t2, ...) -> Tree with keys from *either* trees
    * difference(t1, t2, ...) -> Tree with keys in T but not any of t1, t2, ...
    * symmetric_difference(t1) -> Tree with keys in either T and t1  but not both
    * is_subset(S) -> True if every element in T is in S
    * is_superset(S) -> True if every element in S is in T
    * is_disjoint(S) ->  True if T has a null intersection with S

    Classmethods

    * from_keys(S[,v]) -> New tree with keys from S and values equal to v.

    """

    def __repr__(self):
        """T.__repr__(...) <==> repr(x)"""
        tpl = "%s({%s})" % (self.__class__.__name__, "%s")
        return tpl % ", ".join(("%r: %r" % item for item in self.items()))

    def copy(self):
        """T.copy() -> get a shallow copy of T."""
        tree = self.__class__()
        self.foreach(tree.insert, order=-1)
        return tree

    __copy__ = copy

    def __contains__(self, key):
        """k in T -> True if T has a key k, else False"""
        try:
            self.get_value(key)
            return True
        except KeyError:
            return False

    def __len__(self):
        """T.__len__() <==> len(x)"""
        return self.count

    def __min__(self):
        """T.__min__() <==> min(x)"""
        return self.min_item()

    def __max__(self):
        """T.__max__() <==> max(x)"""
        return self.max_item()

    def __and__(self, other):
        """T.__and__(other) <==> self & other"""
        return self.intersection(other)

    def __or__(self, other):
        """T.__or__(other) <==> self | other"""
        return self.union(other)

    def __sub__(self, other):
        """T.__sub__(other) <==> self - other"""
        return self.difference(other)

    def __xor__(self, other):
        """T.__xor__(other) <==> self ^ other"""
        return self.symmetric_difference(other)

    def discard(self, key):
        """T.discard(k) -> None, remove k from T, if k is present"""
        try:
            self.remove(key)
        except KeyError:
            pass

    def is_empty(self):
        """T.is_empty() -> False if T contains any items else True"""
        return self.count == 0

    def keys(self, reverse=False):
        """T.keys([reverse]) -> an iterator over the keys of T, in ascending
        order if reverse is True, iterate in descending order, reverse defaults
        to False
        """
        return (item[0] for item in self.iter_items(reverse=reverse))

    __iter__ = keys

    def __reversed__(self):
        return self.keys(reverse=True)

    def values(self, reverse=False):
        """T.values([reverse]) -> an iterator over the values of T, in ascending order
        if reverse is True, iterate in descending order, reverse defaults to False
        """
        return (item[1] for item in self.iter_items(reverse=reverse))

    def items(self, reverse=False):
        """T.items([reverse]) -> an iterator over the (key, value) items of T,
        in ascending order if reverse is True, iterate in descending order,
        reverse defaults to False
        """
        return self.iter_items(reverse=reverse)

    def __getitem__(self, key):
        """T.__getitem__(y) <==> x[y]"""
        if isinstance(key, slice):
            return TreeSlice(self, key.start, key.stop)
        else:
            return self.get_value(key)

    def __setitem__(self, key, value):
        """T.__setitem__(i, y) <==> x[i]=y"""
        if isinstance(key, slice):
            raise ValueError("setslice is not supported")
        self.insert(key, value)

    def __delitem__(self, key):
        """T.__delitem__(y) <==> del x[y]"""
        if isinstance(key, slice):
            self.remove_items(self.key_slice(key.start, key.stop))
        else:
            self.remove(key)

    def remove_items(self, keys):
        """T.remove_items(keys) -> None, remove items by keys"""
        # convert generator to a tuple, because the content of the
        # tree will be modified!
        for key in tuple(keys):
            self.remove(key)

    def key_slice(self, start_key, end_key, reverse=False):
        """T.key_slice(start_key, end_key) -> key iterator:
        start_key <= key < end_key.

        Yields keys in ascending order if reverse is False else in descending order.
        """
        return (k for k, v in self.iter_items(start_key, end_key, reverse=reverse))

    def value_slice(self, start_key, end_key, reverse=False):
        """T.value_slice(start_key, end_key) -> value iterator:
        start_key <= key < end_key.

        Yields values in ascending key order if reverse is False else in descending key order.
        """
        return (v for k, v in self.iter_items(start_key, end_key, reverse=reverse))

    def item_slice(self, start_key, end_key, reverse=False):
        """T.item_slice(start_key, end_key) -> item iterator:
        start_key <= key < end_key.

        Yields items in ascending key order if reverse is False else in descending key order.
        """
        return self.iter_items(start_key, end_key, reverse)

    def __getstate__(self):
        return dict(self.items())

    def __setstate__(self, state):
        # note for myself: this is called like __init__, so don't use clear()
        # to remove existing data!
        self._root = None
        self._count = 0
        self.update(state)

    def set_default(self, key, default=None):
        """T.set_default(k[,d]) -> T.get(k,d), also set T[k]=d if k not in T"""
        try:
            return self.get_value(key)
        except KeyError:
            self.insert(key, default)
            return default

    setdefault = set_default  # for compatibility to dict()

    def update(self, *args):
        """T.update(E) -> None. Update T from E : for (k, v) in E: T[k] = v"""
        for items in args:
            try:
                generator = items.items()
            except AttributeError:
                generator = iter(items)

            for key, value in generator:
                self.insert(key, value)

    @classmethod
    def from_keys(cls, iterable, value=None):
        """T.from_keys(S[,v]) -> New tree with keys from S and values equal to v."""
        tree = cls()
        for key in iterable:
            tree.insert(key, value)
        return tree

    fromkeys = from_keys  # for compatibility to dict()

    def get(self, key, default=None):
        """T.get(k[,d]) -> T[k] if k in T, else d.  d defaults to None."""
        try:
            return self.get_value(key)
        except KeyError:
            return default

    def pop(self, key, *args):
        """T.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        if len(args) > 1:
            raise TypeError(
                "pop expected at most 2 arguments, got %d" % (1 + len(args))
            )
        try:
            value = self.get_value(key)
            self.remove(key)
            return value
        except KeyError:
            if len(args) == 0:
                raise
            else:
                return args[0]

    def prev_key(self, key):
        """Get predecessor to key, raises KeyError if key is min key
        or key does not exist.
        """
        return self.prev_item(key)[0]

    def succ_key(self, key):
        """Get successor to key, raises KeyError if key is max key
        or key does not exist.
        """
        return self.succ_item(key)[0]

    def floor_key(self, key):
        """Get the greatest key less than or equal to the given key, raises
        KeyError if there is no such key.
        """
        return self.floor_item(key)[0]

    def ceiling_key(self, key):
        """Get the smallest key greater than or equal to the given key, raises
        KeyError if there is no such key.
        """
        return self.ceiling_item(key)[0]

    def pop_min(self):
        """T.pop_min() -> (k, v), remove item with minimum key, raise ValueError
        if T is empty.
        """
        item = self.min_item()
        self.remove(item[0])
        return item

    def pop_max(self):
        """T.pop_max() -> (k, v), remove item with maximum key, raise ValueError
        if T is empty.
        """
        item = self.max_item()
        self.remove(item[0])
        return item

    def min_key(self):
        """Get min key of tree, raises ValueError if tree is empty."""
        return self.min_item()[0]

    def max_key(self):
        """Get max key of tree, raises ValueError if tree is empty."""
        return self.max_item()[0]

    def nsmallest(self, n, pop=False):
        """T.nsmallest(n) -> get list of n smallest items (k, v).
        If pop is True, remove items from T.
        """
        if pop:
            return [self.pop_min() for _ in range(min(len(self), n))]
        else:
            items = self.items()
            return [next(items) for _ in range(min(len(self), n))]

    def nlargest(self, n, pop=False):
        """T.nlargest(n) -> get list of n largest items (k, v).
        If pop is True remove items from T.
        """
        if pop:
            return [self.pop_max() for _ in range(min(len(self), n))]
        else:
            items = self.items(reverse=True)
            return [next(items) for _ in range(min(len(self), n))]

    def intersection(self, *trees):
        """T.intersection(t1, t2, ...) -> Tree, with keys *common* to all trees"""
        thiskeys = frozenset(self.keys())
        sets = _build_sets(trees)
        rkeys = thiskeys.intersection(*sets)
        return self.__class__(((key, self.get(key)) for key in rkeys))

    def union(self, *trees):
        """T.union(t1, t2, ...) -> Tree with keys from *either* trees"""
        thiskeys = frozenset(self.keys())
        rkeys = thiskeys.union(*_build_sets(trees))
        all_trees = [self]
        all_trees.extend(trees)
        return self.__class__(((key, _multi_tree_get(all_trees, key)) for key in rkeys))

    def difference(self, *trees):
        """T.difference(t1, t2, ...) -> Tree with keys in T but not any of t1,
        t2, ...
        """
        thiskeys = frozenset(self.keys())
        rkeys = thiskeys.difference(*_build_sets(trees))
        return self.__class__(((key, self.get(key)) for key in rkeys))

    def symmetric_difference(self, tree):
        """T.symmetric_difference(t1) -> Tree with keys in either T and t1 but
        not both
        """
        thiskeys = frozenset(self.keys())
        rkeys = thiskeys.symmetric_difference(frozenset(tree.keys()))
        all_trees = [self, tree]
        return self.__class__(((key, _multi_tree_get(all_trees, key)) for key in rkeys))

    def is_subset(self, tree):
        """T.issubset(tree) -> True if every element in x is in tree"""
        thiskeys = frozenset(self.keys())
        return thiskeys.issubset(frozenset(tree.keys()))

    issubset = is_subset  # for compatibility to set()

    def is_superset(self, tree):
        """T.issubset(tree) -> True if every element in tree is in x"""
        thiskeys = frozenset(self.keys())
        return thiskeys.issuperset(frozenset(tree.keys()))

    issuperset = is_superset  # for compatibility to set()

    def is_disjoint(self, tree):
        """T.isdisjoint(S) ->  True if x has a null intersection with tree"""
        thiskeys = frozenset(self.keys())
        return thiskeys.isdisjoint(frozenset(tree.keys()))

    isdisjoint = is_disjoint  # for compatibility to set()


def _build_sets(trees):
    return [frozenset(tree.keys()) for tree in trees]


def _multi_tree_get(trees, key):
    for tree in trees:
        try:
            return tree[key]
        except KeyError:
            pass
    raise KeyError(key)


class CPYTHON_ABCTree(_ABCTree):
    """Base class for the Python implementation of trees.

    T has to implement following methods
    ------------------------------------

    insert(...)
        insert(key, value) <==> T[key] = value, insert key into T

    remove(...)
        remove(key) <==> del T[key], remove key from T

    Properties defined here
    --------------------

    * count -> get item count of tree

    Methods defined here
    --------------------
    * __init__() Tree initializer
    * get_value(key) -> returns value for key
    * clear() -> None.  Remove all items from tree.
    * iter_items(start_key, end_key, [reverse]) -> iterate over all items, yielding (k, v) tuple
    * foreach(f, [order]) -> visit all nodes of tree and call f(k, v) for each node, O(n)
    * pop_item() -> (k, v), remove and return some (key, value)
    * min_item() -> get smallest (key, value) pair of T, O(log(n))
    * max_item() -> get largest (key, value) pair of T, O(log(n))
    * prev_item(key) -> get (k, v) pair, where k is predecessor to key, O(log(n))
    * succ_item(key) -> get (k,v) pair as a 2-tuple, where k is successor to key, O(log(n))
    * floor_item(key) -> get (k, v) pair, where k is the greatest key less than or equal to key, O(log(n))
    * ceiling_item(key) -> get (k, v) pair, where k is the smallest key greater than or equal to key, O(log(n))
    """

    def __init__(self, items=None):
        """T.__init__(...) initializes T; see T.__class__.__doc__ for signature"""
        self._root = None
        self._count = 0
        if items is not None:
            self.update(items)

    def clear(self):
        """T.clear() -> None.  Remove all items from T."""

        def _clear(node):
            if node is not None:
                _clear(node.left)
                _clear(node.right)
                node.free()

        _clear(self._root)
        self._count = 0
        self._root = None

    @property
    def count(self):
        """Get items count."""
        return self._count

    def get_value(self, key):
        node = self._root
        while node is not None:
            if key == node.key:
                return node.value
            elif key < node.key:
                node = node.left
            else:
                node = node.right
        raise KeyError(str(key))

    def pop_item(self):
        """T.pop_item() -> (k, v), remove and return some (key, value) pair as a
        2-tuple; but raise KeyError if T is empty.
        """
        if self.is_empty():
            raise KeyError("pop_item(): tree is empty")
        node = self._root
        while True:
            if node.left is not None:
                node = node.left
            elif node.right is not None:
                node = node.right
            else:
                break
        key = node.key
        value = node.value
        self.remove(key)
        return key, value

    popitem = pop_item  # for compatibility  to dict()

    def foreach(self, func, order=0):
        """Visit all tree nodes and process key, value.

        parm func: function(key, value)
        param int order: inorder = 0, preorder = -1, postorder = +1
        """

        def _traverse(node):
            if order == -1:
                func(node.key, node.value)
            if node.left is not None:
                _traverse(node.left)
            if order == 0:
                func(node.key, node.value)
            if node.right is not None:
                _traverse(node.right)
            if order == +1:
                func(node.key, node.value)

        _traverse(self._root)

    def min_item(self):
        """Get item with min key of tree, raises ValueError if tree is empty."""
        if self.is_empty():
            raise ValueError("Tree is empty")
        node = self._root
        while node.left is not None:
            node = node.left
        return node.key, node.value

    def max_item(self):
        """Get item with max key of tree, raises ValueError if tree is empty."""
        if self.is_empty():
            raise ValueError("Tree is empty")
        node = self._root
        while node.right is not None:
            node = node.right
        return node.key, node.value

    def succ_item(self, key):
        """Get successor (k,v) pair of key, raises KeyError if key is max key
        or key does not exist. optimized for pypy.
        """
        # removed graingets version, because it was little slower on CPython and much slower on pypy
        # this version runs about 4x faster with pypy than the Cython version
        # Note: Code sharing of succ_item() and ceiling_item() is possible, but has always a speed penalty.
        node = self._root
        succ_node = None
        while node is not None:
            if key == node.key:
                break
            elif key < node.key:
                if (succ_node is None) or (node.key < succ_node.key):
                    succ_node = node
                node = node.left
            else:
                node = node.right

        if node is None:  # stay at dead end
            raise KeyError(str(key))
        # found node of key
        if node.right is not None:
            # find smallest node of right subtree
            node = node.right
            while node.left is not None:
                node = node.left
            if succ_node is None:
                succ_node = node
            elif node.key < succ_node.key:
                succ_node = node
        elif succ_node is None:  # given key is biggest in tree
            raise KeyError(str(key))
        return succ_node.key, succ_node.value

    def prev_item(self, key):
        """Get predecessor (k,v) pair of key, raises KeyError if key is min key
        or key does not exist. optimized for pypy.
        """
        # removed graingets version, because it was little slower on CPython and much slower on pypy
        # this version runs about 4x faster with pypy than the Cython version
        # Note: Code sharing of prev_item() and floor_item() is possible, but has always a speed penalty.
        node = self._root
        prev_node = None

        while node is not None:
            if key == node.key:
                break
            elif key < node.key:
                node = node.left
            else:
                if (prev_node is None) or (node.key > prev_node.key):
                    prev_node = node
                node = node.right

        if node is None:  # stay at dead end (None)
            raise KeyError(str(key))
        # found node of key
        if node.left is not None:
            # find biggest node of left subtree
            node = node.left
            while node.right is not None:
                node = node.right
            if prev_node is None:
                prev_node = node
            elif node.key > prev_node.key:
                prev_node = node
        elif prev_node is None:  # given key is smallest in tree
            raise KeyError(str(key))
        return prev_node.key, prev_node.value

    def floor_item(self, key):
        """Get the element (k,v) pair associated with the greatest key less
        than or equal to the given key, raises KeyError if there is no such key.
        """
        # Note: Code sharing of prev_item() and floor_item() is possible, but has always a speed penalty.
        node = self._root
        prev_node = None
        while node is not None:
            if key == node.key:
                return node.key, node.value
            elif key < node.key:
                node = node.left
            else:
                if (prev_node is None) or (node.key > prev_node.key):
                    prev_node = node
                node = node.right
        # node must be None here
        if prev_node:
            return prev_node.key, prev_node.value
        raise KeyError(str(key))

    def ceiling_item(self, key):
        """Get the element (k,v) pair associated with the smallest key greater
        than or equal to the given key, raises KeyError if there is no such key.
        """
        # Note: Code sharing of succ_item() and ceiling_item() is possible, but has always a speed penalty.
        node = self._root
        succ_node = None
        while node is not None:
            if key == node.key:
                return node.key, node.value
            elif key > node.key:
                node = node.right
            else:
                if (succ_node is None) or (node.key < succ_node.key):
                    succ_node = node
                node = node.left
            # node must be None here
        if succ_node:
            return succ_node.key, succ_node.value
        raise KeyError(str(key))

    def iter_items(self, start_key=None, end_key=None, reverse=False):
        """Iterates over the (key, value) items of the associated tree,
        in ascending order if reverse is True, iterate in descending order,
        reverse defaults to False"""
        # optimized iterator (reduced method calls) - faster on CPython but slower on pypy

        if self.is_empty():
            return []
        if reverse:
            return self._iter_items_backward(start_key, end_key)
        else:
            return self._iter_items_forward(start_key, end_key)

    def _iter_items_forward(self, start_key=None, end_key=None):
        for item in self._iter_items(
            left=attrgetter("left"),
            right=attrgetter("right"),
            start_key=start_key,
            end_key=end_key,
        ):
            yield item

    def _iter_items_backward(self, start_key=None, end_key=None):
        for item in self._iter_items(
            left=attrgetter("right"),
            right=attrgetter("left"),
            start_key=start_key,
            end_key=end_key,
        ):
            yield item

    def _iter_items(
        self,
        left=attrgetter("left"),
        right=attrgetter("right"),
        start_key=None,
        end_key=None,
    ):
        node = self._root
        stack = []
        go_left = True
        in_range = self._get_in_range_func(start_key, end_key)

        while True:
            if left(node) is not None and go_left:
                stack.append(node)
                node = left(node)
            else:
                if in_range(node.key):
                    yield node.key, node.value
                if right(node) is not None:
                    node = right(node)
                    go_left = True
                else:
                    if not len(stack):
                        return  # all done
                    node = stack.pop()
                    go_left = False

    def _get_in_range_func(self, start_key, end_key):
        if start_key is None and end_key is None:
            return lambda x: True
        else:
            if start_key is None:
                start_key = self.min_key()
            if end_key is None:
                return lambda x: x >= start_key
            else:
                return lambda x: start_key <= x < end_key


class PYPY_ABCTree(CPYTHON_ABCTree):
    def iter_items(self, start_key=None, end_key=None, reverse=False):
        """Iterates over the (key, value) items of the associated tree,
        in ascending order if reverse is True, iterate in descending order,
        reverse defaults to False"""
        # optimized for pypy, but slower on CPython
        if self.is_empty():
            return
        direction = 1 if reverse else 0
        other = 1 - direction
        go_down = True
        stack = []
        node = self._root
        in_range = self._get_in_range_func(start_key, end_key)

        while True:
            if node[direction] is not None and go_down:
                stack.append(node)
                node = node[direction]
            else:
                if in_range(node.key):
                    yield node.key, node.value
                if node[other] is not None:
                    node = node[other]
                    go_down = True
                else:
                    if not len(stack):
                        return  # all done
                    node = stack.pop()
                    go_down = False


if PYPY:
    ABCTree = PYPY_ABCTree
else:
    ABCTree = CPYTHON_ABCTree



import numpy as np
import pyarrow as pa
import json, time


def json_to_struct(arr):
    return pa.array(np.vectorize(json.loads)(arr))


def text_extract(text, key):
    value = '"' + key + '":'
    start = text.find(value)
    if start > -1:
        end = text[start:].find(("," if "," in text[start:] else "}"))
        return eval(text[start + len(value) : start + end])
    else:
        return None


def json_extract(arr, key):
    f = np.vectorize(lambda t: text_extract(t, key))
    return pa.array(f(arr.to_numpy()))



"""
Limit Node

This is a SQL Query Execution Plan Node.

This Node returns up to a specified number of tuples.
"""
from typing import Iterable
from pyarrow import concat_tables
from opteryx.engine.query_statistics import QueryStatistics
from opteryx.engine.planner.operations.base_plan_node import BasePlanNode


class LimitNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        self._limit = config.get("limit")

    def execute(self, data_pages: Iterable) -> Iterable:

        result_set = []
        row_count = 0

        for page in data_pages:
            row_count += page.num_rows
            result_set.append(page)
            if row_count > self._limit:
                break

        yield concat_tables(result_set).slice(0, self._limit)



"""
Dataset Reader Node

This is a SQL Query Execution Plan Node.

This Node reads and parses the data from a dataset into a Table.

We plan to do the following:

USE THE ZONEMAP:
- Respond to simple aggregations using the zonemap, such as COUNT(*)
- Use BRIN and selections to filter out blobs from being read that don't contain
  records which can match the selections.

PARALLELIZE READING:
- As one blob is read, the next is immediately cached for reading
"""
import time
from enum import Enum
from typing import Iterable
from opteryx.engine.planner.operations import BasePlanNode
from opteryx.engine import QueryStatistics
from opteryx.storage import file_decoders
from opteryx.storage.adapters import DiskStorage
from opteryx.storage.schemes import MabelPartitionScheme


class EXTENSION_TYPE(str, Enum):
    # labels for the file extentions
    DATA = "DATA"
    CONTROL = "CONTROL"


do_nothing = lambda x: x

KNOWN_EXTENSIONS = {
    "complete": (do_nothing, EXTENSION_TYPE.CONTROL),
    "ignore": (do_nothing, EXTENSION_TYPE.CONTROL),
    "metadata": (do_nothing, EXTENSION_TYPE.CONTROL),
    "arrow": (file_decoders.arrow_decoder, EXTENSION_TYPE.DATA),
    "jsonl": (file_decoders.jsonl_decoder, EXTENSION_TYPE.DATA),
    "orc": (file_decoders.orc_decoder, EXTENSION_TYPE.DATA),
    "parquet": (file_decoders.parquet_decoder, EXTENSION_TYPE.DATA),
    "zstd": (file_decoders.zstd_decoder, EXTENSION_TYPE.DATA),
}


class DatasetReaderNode(BasePlanNode):
    def __init__(self, statistics: QueryStatistics, **config):
        """
        The Dataset Reader Node is responsible for reading the relevant blobs
        and returning a Table/Relation.
        """
        self._dataset = config.get("dataset", "").replace(".", "/") + "/"
        self._reader = config.get("reader", DiskStorage())
        self._partition_scheme = config.get("partition_scheme", MabelPartitionScheme())

        self._statistics = statistics

        # pushed down projection
        self._projection = config.get("projection")
        # pushed down selection
        self._selection = config.get("selection")

    def __repr__(self):
        return self._dataset

    def execute(self, data_pages: Iterable) -> Iterable:

        if self._dataset.lower() == "$satellites/":
            from opteryx import samples

            yield samples.satellites()
            return
        if self._dataset.lower() == "$planets/":
            from opteryx import samples

            yield samples.planets()
            return

        partitions = self._reader.get_partitions(
            dataset=self._dataset,
            partitioning=self._partition_scheme.partition_format(),
        )

        self._statistics.partitions_found = len(partitions)

        for partition in partitions:

            self._statistics.partitions_scanned += 1

            # Get a list of all of the blobs in the partition.
            blob_list = self._reader.get_blob_list(partition)

            # remove folders, end with '/'
            blob_list = [blob for blob in blob_list if not blob.endswith("/")]

            # Track how many blobs we found
            self._statistics.count_blobs_found += len(blob_list)

            # Filter the blob list to just the frame we're interested in
            if self._partition_scheme is not None:
                blob_list = self._partition_scheme.filter_blobs(blob_list)

            # If there's a zonemap for the partition, read it
            #        zonemap = {}
            #        zonemap_files = [blob for blob in blob_list if blob.endswith("/frame.metadata")]
            #        if len(zonemap_files) == 1:
            #            # read the zone map into a dictionary
            #            try:
            #                import orjson
            #                zonemap = orjson.loads(self._reader.read_blob(zonemap_files[0]))
            #            except:
            #                pass

            if len(blob_list) > 0:
                self._statistics.partitions_read += 1

            for blob_name in blob_list:

                # the the blob filename extension
                extension = blob_name.split(".")[-1]

                # find out how to read this blob
                decoder, file_type = KNOWN_EXTENSIONS.get(extension, (None, None))
                # if it's not a known data file, skip reading it
                if file_type != EXTENSION_TYPE.DATA:
                    continue

                # can we eliminate this blob using the BRIN?
                #            pass

                # we're going to open this blob
                self._statistics.count_data_blobs_read += 1

                start_read = time.time_ns()

                # Read the blob from storage, it's just a stream of bytes at this point
                blob_bytes = self._reader.read_blob(blob_name)

                # record the number of bytes we're reading
                self._statistics.bytes_read_data += blob_bytes.getbuffer().nbytes

                # interpret the raw bytes into entries
                pyarrow_blob = decoder(blob_bytes, self._projection)

                self._statistics.time_data_read += time.time_ns() - start_read

                # we should know the number of entries
                self._statistics.rows_read += pyarrow_blob.num_rows
                self._statistics.bytes_processed_data += pyarrow_blob.nbytes

                # yield this blob
                print(f"reader yielding {blob_name} {pyarrow_blob.shape}")
                yield pyarrow_blob



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
These are a set of functions that can be applied to data.
"""
import orjson
import datetime
import numpy
import pyarrow
from pyarrow import compute
from cityhash import CityHash64

from opteryx.engine.functions.date_functions import *


from opteryx.utils.text import levenshtein_distance


def get_random():
    from opteryx.utils.entropy import random_range

    return random_range(0, 999) / 1000


def concat(*items):
    """
    Turn each item to a string and concatenate the strings together
    """
    sep = ""
    if len(items) == 1 and (
        isinstance(items[0], (list, tuple, set)) or hasattr(items[0], "as_list")
    ):
        items = items[0]
        sep = ", "
    return sep.join(map(str, items))


def to_string(val):
    if isinstance(val, (list, tuple, set)):
        return concat(val)
    if hasattr(val, "mini"):
        return "\\" + val.mini.decode("UTF8") + "\\"
    if isinstance(val, dict):
        return "\\" + orjson.dumps(val).decode("UTF8") + "\\"
    else:
        return str(val)


def parse_number(parser, coerce):
    def inner(val):
        if val is None:
            return None
        return coerce(parser(val))

    return inner


def get_md5(item):
    # this is slow but expected to not have a lot of use
    import hashlib
    return hashlib.md5(str(item).encode()).hexdigest()  # nosec - meant to be MD5

def attempt(func):
    try:
        return func()
    except:
        return None

def not_implemented(*args):
    raise NotImplementedError()


def _vectorize_single_parameter(func):
    def _inner(array):
        for a in array:
            yield [func(a)] 
    return _inner

def _vectorize_double_parameter(func):
    def _inner(array, p1):
        for a in array:
            yield [func(a, p1)]
    return _inner 


FUNCTIONS = {

    # STRINGS
    # VECTORIZED
    "LENGTH": compute.utf8_length, # LENGTH(str) -> int
    "UPPER": compute.utf8_upper, # UPPER(str) -> str
    "LOWER": compute.utf8_lower, # LOWER(str) -> str
    "TRIM": compute.utf8_trim_whitespace, # TRIM(str) -> str

    # STRINGS
    # LOOPED FUNCTIONS
    "LEFT": _vectorize_double_parameter(lambda x, y: str(x)[: int(y)]),
    "RIGHT": _vectorize_double_parameter(lambda x, y: str(x)[-int(y) :]),

    # HASHING & ENCODING
    "HASH": _vectorize_single_parameter(lambda x: format(CityHash64(str(x)), "X")),
    "MD5": _vectorize_single_parameter(get_md5),

    # NUMERIC
    "ROUND": compute.round,
    "FLOOR": compute.floor,
    "CEIL": compute.ceil,
    "ABS": compute.abs,
    "TRUNC": compute.trunc,

    # NOT CONVERTED YET






    "RANDOM": get_random,  # return a random number 0-99

    # DATES & TIMES
    "YEAR": get_year,
    "MONTH": get_month,
    "MONTH_NAME": not_implemented,  # the name of the month
    "DAY": get_day,
    "DAY_NAME": not_implemented,  # the name of the day
    "DATE": numpy.datetime64, # this should be vectorized
    "QUARTER_OF_YEAR": get_quarter,
    "WEEK_OF_YEAR": get_week,
    "DAY_OF_YEAR": not_implemented,  # get the day of the year
    "DAY_OF_WEEK": not_implemented,  # get the day of the week (Monday = 1)
    "HOUR": get_hour,
    "MINUTE": get_minute,
    "SECOND": get_second,
    "TIME": get_time,
    "CURRENT_DATE": datetime.date.today,
    "TODAY": not_implemented,
    "YESTERDAY": not_implemented,
    "NOW": datetime.datetime.now,
    "DATE_ADD": not_implemented,  # date, number, part
    "DATE_DIFF": not_implemented,  # start, end, part
    "AGE": not_implemented,  # 8 years, 3 months, 3 days
    "FROM_EPOCH": not_implemented,  # timestamp from linux epoch formatted time
    "TO_EPOCH": not_implemented,  # timestamp in linux epoch format
    "DATE_PART": not_implemented,  # DATE_PART("YEAR", timestamp)
    "TIMESTAMP": not_implemented,  # parse input as a TIMESTAMP

    
    "STRING": to_string,
    "VARCHAR": to_string,
    "MID": lambda x, y, z: str(x)[int(y) :][: int(z)],
    "CONCAT": concat,
    "LEVENSHTEIN": levenshtein_distance,
    "REGEXP_MATCH": not_implemented,  # string, pattern -> boolean
    "REPLACE": not_implemented,  # string, pattern to find, pattern to replace -> string
    # NUMBERS
    "INTEGER": parse_number(float, int),
    "DOUBLE": parse_number(float, float),
    # BOOLEAN
    "BOOLEAN": lambda x: str(x).upper() != "FALSE",
    "ISNONE": lambda x: x is None,

    # OTHER

    "SORT": sorted,
    "TRY": attempt,
    "LEAST": min,
    "GREATEST": max,
    "UUID": not_implemented,  # cast value as UUID
    "GET": not_implemented,  # GET(LIST, index) => LIST[index] or GET(STRUCT, accessor) => STRUCT[accessor]
}


def is_function(name):
    return name.upper() in FUNCTIONS



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .version import __version__
from opteryx.connection import Connection


apilevel = "1.0"
threadsafety = 0
paramstyle = "format"


def connect(*args, **kwargs):
    return Connection(*args, **kwargs)


try:
    import dotenv  # type:ignore
except ImportError:  # pragma: no cover
    dotenv = None  # type:ignore

from pathlib import Path

env_path = Path(".") / ".env"

#  deepcode ignore PythonSameEvalBinaryExpressiontrue: false +ve, values can be different
if env_path.exists() and (dotenv is None):  # pragma: no cover  # nosemgrep
    # using logger here will tie us in knots
    print("`.env` file exists but `dotEnv` not installed.")
elif dotenv is not None:  # pragma: no cover
    dotenv.load_dotenv(dotenv_path=env_path)



class FlowRunner:
    def __init__(self, flow):
        self.flow = flow
        self.cycles = 0

    def __call__(
        self, data: dict = {}, context: dict = None, trace_sample_rate: float = 1 / 1000
    ):
        """
        Create a `run` of a flow and execute with a specific data object.

        Parameters:
            data: dictionary, any (optional)
                The data the flow is to process, opinionated to be a dictionary
                however, any data type is accepted.
            context: dictionary (optional)
                Additional information to support the processing of the data
            trace_sample_rate: float (optional)
                The sample for for to emit trace messages for, default is
                1/1000.
        """
        if not context:
            context = {}

        # create a run_id for the message if it doesn't already have one
        if not context.get("run_id"):
            context["run_id"] = str(entropy.random_string(32))

        # create a tracer for the message
        if not context.get("execution_trace"):
            context["execution_trace"] = TraceBlocks(uuid=context["run_id"])

        # if trace hasn't been explicitly set - randomly select based on a sample rate
        if not context.get("trace") and trace_sample_rate:
            context["trace"] = (
                entropy.random_range(1, round(1 / trace_sample_rate)) == 1
            )  # nosec

        try:
            # start the flow, walk from the nodes with no incoming links
            for operator_name in self.flow.get_entry_points():
                self._inner_runner(
                    operator_name=operator_name, data=data, context=context
                )
        except TimeExceeded as te:
            raise te
        except (Exception, SystemExit) as err:
            if hasattr(self, "error_writer"):
                error_log_reference = "NOT LOGGED"
                try:
                    error_payload = (
                        f"timestamp  : {datetime.datetime.today().isoformat()}\n"
                        f"location   : flow_runner\n"
                        f"error type : {type(err).__name__}\n"
                        f"details    : {err}\n"
                        "========================================================================================================================\n"
                        f"{wrap_text(render_error_stack(), 120)}\n"
                        "=======================================================  context  ======================================================\n"
                        f"{wrap_text(str(context), 120)}\n"
                        "========================================================  data  ========================================================\n"
                        f"{wrap_text(str(data), 120)}\n"
                        "========================================================================================================================\n"
                    )
                    error_log_reference = self.error_writer(  # type:ignore
                        error_payload
                    )  # type:ignore
                except:
                    # if we have a uncaught failure, make sure it's logged
                    get_logger().alert(  # type:ignore
                        f"FLOW ABEND - {type(err).__name__} - {err} ({error_log_reference})"
                    )
            raise err

        # if being traced, send the trace to the trace writer
        if context.get("trace", False) and hasattr(self, "trace_writer"):
            self.trace_writer(  # type:ignore
                context["execution_trace"], id_=str(context.get("run_id"))
            )

    def _inner_runner(
        self, operator_name: str = None, data: dict = {}, context: dict = None
    ):
        """
        Walk the dag/flow by:
        - Getting the function of the current node
        - Execute the function, wrapped in the base class
        - Find the next step by finding outgoing edges
        - Call this method for the next step
        """
        self.cycles += 1
        if not context:
            context = {}

        operator = self.flow.get_operator(operator_name)
        if operator is None:
            raise FlowError(f"Invalid Flow - Operator {operator_name} is invalid")
        if not hasattr(operator, "error_writer") and hasattr(self, "error_writer"):
            operator.error_writer = self.error_writer  # type:ignore
        out_going_links = self.flow.get_outgoing_links(operator_name)

        outcome = operator(data, context)

        if outcome:
            if not type(outcome).__name__ in ["generator", "list"]:
                outcome_data, outcome_context = outcome
                outcome = [(outcome_data, outcome_context)]
            for outcome_data, outcome_context in outcome:
                for operator_name in out_going_links:
                    self._inner_runner(
                        operator_name=operator_name,
                        data=outcome_data,
                        context=outcome_context.copy(),
                    )



"""
planets
---------

This is a sample dataset build into the engine, this simplifies a few things:

- We can write test scripts using this data, knowing that it will always be available. 
- We can write examples using this data, knowing the results will always match.

This data was obtained from:
https://github.com/devstronomy/nasa-data-scraper/blob/master/data/json/planets.json

Licence @ 02-JAN-2022 when copied - MIT Licences attested, but data appears to be
from NASA, which is Public Domain.

To access this dataset you can either run a query against dataset :planets: or you
can instantiate a PlanetData() class and use it like a Relation.

This has a companion dataset, $satellites, to help test joins.
"""

if __name__ == "__main__":
    import sys
    import os

    sys.path.insert(1, os.path.join(sys.path[0], "../.."))

import io
import base64
import pyarrow.parquet as pq  # type:ignore


class PlanetData:
    @staticmethod
    def get():

        return pq.read_table(
            io.BytesIO(
                base64.b85decode(
                    b"P(e~L6$BNK0Tob86%rK$5&!@wwJ-f3Nc{i+zySaP00RI700jUB00#gF00{s90000000>k900RLP02Lq=GAtDm6$BLq6$Ts_2nhfH0000002l}X000000000M04N9v000000000O2mt^90000000000D77#BAP|8700IC201^QN1rRb+b_f6f04AaW92Eo^H3R?#85ja-WEBh+5*Ek;7McPkb|wfM7zhaf00000000;W0RR9100000762#+2><{9000007zhCX000000000PEENP51Qh}R6#x|k6#@VN02KrkumKgS0ZbJV6$BCh04TLD{UBIn0su0_5Ey{PEe60M>$~hy2qbo7@W~i7L34^|5h()QfbpBdRdvOpCGsE4RXhHZrvSCYk5*dxg+<eb->k(;ely5b^3sn20RV##R1_5e6(AKdEEN(J1QiAq1{^j3C<Rt!Zgq1Q1w~<UbZ7to04TLD{U8v500062000sJ1O*T>RCWje001VO1{@U(88rj|1{oLxZeeX@6$};<7PbNw!U86w1t!`995w(b1y*Hlb#oX6MPYJuXaE^36$BLo6#@Vi02Krk0ssI26$BNK0Tq@3OcfFp1QGxMD77#BAV?Jg0E7b{h3iqc9@Rg46s||%dJ{n5dK9il;TJ$927#4Bi{w85sGUFnz=1$BvsFL<fLTDFJS^0j=8ZoD6aWE)kURku02Lq=GAtDm6$BLq6$Ts_2mk;8005|+Ko|(0JS^0j=8ZoV04N9m00000sGUF<2%kJG)SBjvKL7v#D77#BAP|8700IC201^QN1rRb+b_f6f049nG92E)~H3R?#85jg@VRLg83>Fd=$^sVR0w&%FCa4D-7zh9W0001}oj@1}pFAwon&yo^762#+000000H~cn7zm#{EYzCjjXwYxEENP51Qh}R6#x|k6#@VN02KrkkO38bOcfFp1QGxMD77#BAV{?V05k#s0000053@i3q{l!2>c~I<h`K;@0YQLm{XoF==m0N-qCf}$ATj6xmViZ81cM+TNC1u%02Lq=GAtDm6$BLq6$Ts_2mk;806=sBK^O=C00000grYze04N9m0000$bOAva2mk;8004xdKmY&$D77#BAP|8700IC201^QN1rRb+b_f6f049<S92E)~H3R?#85js;X<=<;bY*fC3>Fd=$O0C~0w(4RCdvyO7zh9W002OA0YMlD000000ED7I762#+00000Ky(2?7zh9W0001lqCfx{EENP51Qh}R6#x|k6#@VN02KrkkO38LOcfFp1QGxMD77#BAV_rq0N4Nk0000pwLk!Sv_Jrgx~{mCc!k`Pn3-OnKnMU}U@$Nk7z_*s&P)a@Ti^&402Lq=GAtDm6$BLq6$Ts_2mk;8004@$Ko|%B00000c!fX~04N9m00000inTx(2mk;8004M}KmY&$D77#BAP|8700IC201^QN1rRb+b_f6f049<W92E)~H3R?#85jp-Wo~n6ba@pF77`Z70v5OeCgu<(+7BEU2mk;8004@$Ko|%B00000c!fX~04N9m00000inTx(2mk;8004M}KmZvm6$BLo6#@Vi02Krk0ssI26$BN40Tpgc6%Z8!5&!@wwJ-f3Ky?8C@S2&KnVFdlK+Vj|%*@OoKqDD9KmZ~@W+6ZTCgwi|05em7Hh>J=rIm5bFqaho6&w{VEEN(J1QiAq1{@d&nwgoInHe`g7zk!&W@ct)=06qyC<vOFnVFdxH$WH&W@ct)W@hF;0000ewJ-f34sieg0ssI25&;7Rh%nX%001V87914{88rj|1{oL!XL4b7X>@rN3>Fd=v;r2i0w&}XCfE}k7zmn~nVFdxH$WH&W@ct)W@hF;762#+nwgoInHe`g7zk!&W@ct)=05-#EENP51Qh}R6#x|k6#@VN02KrkkO38ROcfFp1QGxMD77#BAV`e?0QfUAGcz+Y5kSq%%*@QpBtT{+KmZiLO+idGfH%zaKL-F{;F$xL8VuyNG6LM#f+ZCI6(AKdEEN(J1QiAq1{@d&000000KiQ^7zoYG%*@Qp^gk8=C<p)m0002MO+Xk3&CJZq%*^yZ0000ewJ-f35P<*y0ssI25&;AS5HeJD2mk;8CXgB&6$%+O1ONsZ7!GA~V_|S*R%L8&V`+4G6$};<7RUk?z5*uX7$(ve92f`y00000z)e6H2+hpQ%*@R6KNbKe2mk;8006*EKo|(k%*@Qp%=AA187vh96$BLm02Kfg1Qh}R000#P6_5cHmH|u^5)}jz001bpFa01$6#@WM17>DsW@i4CKmY&$0D#E0z-Hz*K$@AEnVFe5K+Vj|%*@OqKxQRCGcmv!FhH7_nVFd>W55IeLkR{qfM!>)1Qh@kAQdt!6%rK$6$TXs92f{@W@ct){*^!&2mk;80D#E0z!m@~2xewxW@i4CKo|%B0001h$hN=$001bpFa018fdBvk0000I0R#mQGE{a50000c#vU9M3K=y700tQt4svgFVRUJ4Zct@%X>Vi|3>Fd=$^sVR0w$UrCgd9&7zk!&W@cvol|UE>00000fXKGM762#+W@ct)X8x5x7zh9W004l<w!i=xEENP51Qh}R6#x|k6#@VN02KrkkO38r0ZbJV6$BCh04TLD{UAv70RRL8nwgoInSQW90OX}W060K1Gcz+Yvp7J_%p*W%B{MNVnwgoInHexZnwgoInJHsH1pqrofE&OPXtIJM6#x|=6*4Rp5)}j$1{DSz7zmn~nVFe>us|3H&CJZq%*-P|762#+nwgoInSQW97zoYG%*@QpBR~KE04TLD{U8v500062000sJ1O*T>RCWje001WRA{-S888rj|1{oL&Y-Mg|bZAdzL}7Uq3>Fd=$^sVJ0w%~ICYm4|7zmn~nVFe>us|3H&CJZq%*-P|762#+nwgoInSQW97zoYG%*@QpBR~KdEENP51Qh}R6#x|k6#@VN02KrkkO38t0ZbJV6$BCh04TLD{UAvF0RSWeGcz+YGxJP9&CJZq%nVyVGqYlRY*dIq000000A`kPra+pRnVA`nu|Q^KW@cs*w?G5{V4#(OXAT~KB^3Y_AQdt!6%rK$6$TXs92f{@W@ct)61PAY2s1M?Gc)r{Ko$Tf2xewxW@Zw%Ko|%!Gcz+Y^GrYh001bpFa018fdBvk0000I0R#mQGE{a50000co+lg?3K=y700tQt4`gX`bYX5|WkzyuZBun_6$};<7Rmw^+yW-@B__fo92f{@W@ct)61PAY2s1M?Gc)r{Ko$Tf2xewxW@Zw%Ko|%!Gcz+Y^GrYh87vh96$BLm02Kfg1Qh}R000#P6_5cHkO52;5)}jz001bpFa01${Q&^<0ssI20000-Kmg!cKr>@vKr_>6Kma6%KxQJ9K$@AEnVD*(fLyUa&CJZq%v7;J1^_z+fgm6V2m*ox;0P4}6(AKdEEN(J1QiAq1{@d&00000fLyUa7zh9W00000M?e+;C<p)m004kou|OCI00000002io0000ewJ-f35P<*y0ssI25&;AS5HeJD2mk;8CeSM!6$%+O1ONsZ7z%J@a%pH~Y-w+96$};<7Rmw^+yW+~Dkk(O92f`y0001hT(LkH2mk;80000-Ko$Tf2mk;80DxSvKo|%B0000007pOo87vh96$BLm02Kfg1Qh}R000#P6_5cHkO52;5)}jz001bpFa01${Q&^v0y8r+Gc$8hK$@AEnVC6TKr;hlKxQUyK+Vj|%*=#|Kme+jKr?rzv%#_E&%8he087bOISh;dOTb7AECCe&6(AKdEEN(J1QiAq1{@d&W@ct)=Fhx97zi^nGcz-DQ9u>|C<ta|W@hHkyg(QTGcz+YGjmZu0000ewJ-f35P<*y0ssI25&;AS5HeJD2mk;8CipKL6$%+O1ONsZ7zkl-Xk~0^Z*CO~77`Z90v6l?Cd@7-oGly}2xewxX6DblKo|%!Gcz+Yb5TGR04NA%W@cvQ&%8hw2s1M?Gc$8hKmZvm6$BLo6#@Vi02Krk0ssI26$BNK0Tq5s6%rK$5&!@wwJ-f3NVNd~Gy(ts00000RzPMJY(O*9c0d4lg+Ku7uz>x<K;PW}GwmGqKnMU}2Dky*SP=|@fFK}90FD&^6(AKdEEN(J1QiAq1{@d&00000037x}7zh9W00000RzMa2C<p)m0000S_COd2000000034%0000ewJ-f35P<*y0ssI25&;AS5HeJD2mk;8CWtf~6$%+O1ONsZ7!7Z7Vrg_?Y*1x#X>Vi|3>Fd=$O0C~0w&xtCcrTq7zh9W0000S_COd2000000034%762#+00000037x}7zh9W00000RzLt5EENP51Qh}R6#x|k6#@VN02KrkkO38n0ZbJV6$BCh04TLD{UAu?0RYqjGcz+YGqXoP0DwV2&CJZq%*;JNnwgoInHe}hGb%u4BO93&K+Vj|%*@OZKm`Cp4Vjr2v;hbUz!4Py6(AKdEEN(J1QiAq1{@d&Gcz+YGqXoP7zoYG%*@Qp5<nIJC<rq%Gcz-@M?e?|&CJZq%*+x%0000ewJ-f35P<*y0ssI25&;AS5HeJD2mk;8Ca^dh6$%+O1ONsZ7!Pl9Vrg_?Y*uA#Z)0h6c@+#65*Eq=7S;kLgf}MIH5?cSGcz+YGqXoP7zoYG%*@Qp5<nIJC<rq%Gcz-@M?e?|&CJZq%*+x%02wS51Qi4o0ss{N6$BLm0000L1Qn106?{w;5)}jz001bpFa01$tpNZu0ssI20000SKr;(K0A~I_&CK*a00cmqnd$sLGcz+YGcz$j2LL-$03)CWSOmlcIa+W86#x|=6*4Rp5)}j$1{DSz7zi^nGcz+YF+dmy000000001h762#+Gcz+YGcz$j7zh9W000000Du4h04TLD{U8v500062000sJ1O*T>RCWje001V;JRB7Y88rj|1{oL<Z*pR3bYW~sZewg|Zeet3Z*CO~77`Z70v5&sCZIbehB_P=2s1M?Gcz+WKo|%B00000004j%04NAEGcz+YGciCI2mk;800000fB+dR6$BLo6#@Vi02Krk0ssI26$BNK0TrSFOcfFp1QGxMD77#BAV@(106hxVqi{XSKY8gUF$k=DKb*uMuySaTKdSN4O=1wZKS~S&D~A@TKgo8!pBgW%KU!ov-%pLFKe|Z_0xO4xKM1TGT4X!VKNSEKAQdt!6%rK$6$TXs92f`)tQ=ZoJI_BD2zluyF$k=DKNbKe2nehkT4X!VKNtvk=_WA<tb9KJ001bpFa018fdBvk0000I0R#mQGE{a50000cmO&g93K=y700tQt6K`^2X>?(1MPp-SZgg^KV`+4G6$};<7Rmw^?gA$2KPIR?92f`)tQ=ZoJI_BD2zluyF$k=DKNbKe2nehkT4X!VKNtvk=_WA<tb9KJ87vh96$BLm02Kfg1Qh}R000#P6_5cHgaJ$y5)}jz001bpFa01$%>e*}0-VGkuySajKh4a{%*-riKxQ{UGdV!b%m}kOb689~KmY&$005v~Kn4Ii2VNOCa~OyWZfwC46#x|=6*4Rp5)}j$1{DSz7zoYG%*@OzW<VGSoWvloa%iDH762#+&CJZq%q(U=7zmuiAh2?1p+5is04TLD{U8v500062000sJ1O*T>RCWje001V$MI03h88rj|1{oL-Z(?j|adl~Qc~oyta$;$86$};<7Rmw^&;llyL?-k?92f}A%*@QpEM`C$2%N+quySajKNbKe2+hpQ%*-riKo|&|#2~P8XrVs<87vh96$BLm02Kfg1Qh}R000#P6_5cHXiOCn6$BCh04TLD{UAtP0RX@N00000;AB7meJ(&yz<^s}U}ykoAZ)-00AOG+5DW&gf#v{~mAoxD0u=xiAQdt!6%rK$6$TXs92f`y000000DV9h2mk;8001Ctz!m@~2mk;80001eKo|%B00000AZ)+@001bpFa018fdBvk0000I0R#mQGE{a50000c&Pf~<3K=y700tQt4{c>(Zd7G$aAk5~bairN6$};<7RUk?vH~WeNG6;|92f`y000000DV9h2mk;8001Ctz!m@~2mk;80001eKo|%B00000AZ)+@87vh96$BLm02Kfg1Qh}R000#P6;KsWOce?h1QGxMD77#BASnF+0Js1ES3m&pKYJ9eN8x&eKk;beh_(3SJ^}zh7#zR?6#x|&6)G$h5)}j$1{DSz7zh9W00000S3npD000000001h76>Q^000000037&7zh9W000000Du4h04TLD{U8is00093000FK0RsbwMgS(XO&k>p88rj|1{oL+b9HiNVPj=ba%FRKb#i4D3>Fd={s9({0w$15CZtOo7zh9W00000S3npD000000001h76>Q^000000037&7zh9W000000Du4)EENP51Qh}R6#x|k6#@VN02KrkfB_XuOcf9n1QGxMD77#BAVBQ^0I&c60RRF3PXImu8vqUf1poj5000002UG$80|6BP6&w{VEEN(J1QiAq1{@d&PXGV_000007zh9W00000000&MC<spg00000000;W00000000000000ewJ-f34sieg0ssI25&;7RKs43|001VOQ5+Qn88rj|1{oL)Zgp*9WpYnuO>b{*a}^905*D-q7Lo!c^iL*=PaGHsPXGV_000007zh9W00000000&MC<spg00000000;W000000000002wS51Qi4o0ss{N6$BLm0000L0vY@jH2@d}b7N>_ZDAEC02Kr!0vG~mWB?TmB?1@(ZeeX@B>+qu0000L3MB#<1Z`n+a{v_zB?1@-WNBe-Wprh702K-)0vHEmWo~n6ba?<33MB#<2WN6&c4>5Z02K-)0vHZub7Ns}WmaWuZ)0h6c>om(B?1@@a&L5DbZKvHP-SvyZ)5-!3MB#<3v6X>XLM*!W<+6m02K-)0vHcuX>)X8ZewLea&K)@b#4F^3MB#<3UFm|X=r6^X>V=-6$&K+7zkl-Xk~0^Z*Bk;3MB#<4R3N{X>?(1P-SvyZ)5-!3MB#<4{vf}X>?(1R%L8&V`+4G02K-)0vHl+a$;$8VQfimV{B<|VRUJ4ZU7YuB?1@|Z*pR3bYW~oV`F7*baH89X>@r26$&K+7!Yq_Y-w?IX>@s1Z%=Y!X><S;3MB#<4{c>(Zd7G$aAk5~bairN02K-)0vHc-b#i85V`WfsWpi_Na%BJ&1SJ9(4Q_R9Vr6nqW=(HzZgT(@5*ZvB{1hgl0vr_t88rj|1{oLvX=D`)77`Z70v4JACUzzW92f`*000000000O2mt^90000002Tl!2nhfH0000002l}X000000000087vh96$BLm02Kfg1Qh}R001VO1{@U(88rj|1{oLxZeeX@6$};<7PbNw!U86w1t!`995w(b1y*Hlb#oX6MPYJuXaE^36$BLo6#@Vi02Krk0ssI2CW;Ci6$%+O1ONsZ7zAx$b8{6877`Z90v6%|Cf*1ps0SPv2mk;8005|+Ko|(0JS^0j=8ZoV04N9m00000sGUF<2%kJG)SBjvKL8ml6$BLo6#@Vi02Krk0ssI2CXx;u6$%+O1ONsZ7zku(VQpn}WpWh^77`Z70v5;uCguz#$_pGA2mk;806=sBK^O=C00000grYze04N9m0000$bOAva2mk;8004xdKmZvm6$BLo6#@Vi02Krk0ssI2CXx~y6$%+O1ONsZ7zbo!ZgXjLc@+#65*Ek;7PtZ?<`5>@4;&Z>000000E)Fh7zh9W0001Zg+LYnC<p)m0001rwLlmM000000C<H!02wS51Qi4o0ss{N6$BLm0000cj20Xf3K=y700tQt2WN6&c4>5Z6$};<7PJBuv;rpN6eidc92f|inVFfH88<)}2xewxW@cvQKNbKe2%4FhnVA_kKo|&SW@ct)X68Qt87vh96$BLm02Kfg1Qh}R001VC8XOf088rj|1{oL*WpiU;aAj6yY;R*}ba@pF77`Z70v5gkCgd0<(ia>U2mk;8006*EKo|(k%*@Qp%=AAN04N9m00000z)e6H2+hpQ%*@R6KL8ml6$BLo6#@Vi02Krk0ssI2CdM8d6$%+O1ONsZ7!Go8bYXO9Z*EX!a%pd56$};<7Rmw^;sPd`9VX-(92f{@W@ct){*^!&2mk;80D#E0z!m@~2xewxW@i4CKo|%B0001h$hN=$87vh96$BLm02Kfg1Qh}R001WRA{-S888rj|1{oL&Y-Mg|bZAdzL}7Uq3>Fd=$^sVJ0w%~ICYm4|7zmn~nVFe>us|3H&CJZq%*-P|762#+nwgoInSQW97zoYG%*@QpBR~KdEENP51Qh}R6#x|k6#@VN04AO%92E)~H3R?#85j>_X>)X8ZewLea&K)@b#4_577`Z90v6l?Ch{dF!Xz9R2xewxW@Zw%Ko|%!Gcz+Y^GrY%04NA%W@ct)61PAY2s1M?Gc)r{KmZvm6$BLo6#@Vi02Krk0ssI2CeSM!6$%+O1ONsZ7z%J@a%pH~Y-w+96$};<7Rmw^+yW+~Dkk(O92f`y0001hT(LkH2mk;80000-Ko$Tf2mk;80DxSvKo|%B0000007pOo87vh96$BLm02Kfg1Qh}R001WVFB}yL88rj|1{oL#VQ^?=Y-w+96$};<7Rmw^+yW-dE+(8U92f{@W@cvQ&%8hw2s1M?Gc$8hKo$Tf2xewxX6DblKo|%!Gcz+Yb5TG587vh96$BLm02Kfg1Qh}R001V4G#nKQ88rj|1{oL)Z*pR3bYW~zWpZh6WEBh+5*Ek;7RUl7+%hJ>F&r2O00000037x}7zh9W00000RzMa2C<p)m0000S_COd2000000034%02wS51Qi4o0ss{N6$BLm0000cus9qQ3K=y700tQt4{vf}X>?(1R%L8&V`+4G6$};<7Rmw^)&eGkHzwLO92f{QGcz+YvqwM}2+hpQ%*@OZKo$Tf2s1M?Gc&VCKo|(k%*@Qp%o0EV87vh96$BLm02Kfg1Qh}R001V;JRB7Y88rj|1{oL<Z*pR3bYW~sZewg|Zeet3Z*CO~77`Z70v5&sCZIbehB_P=2s1M?Gcz+WKo|%B00000004j%04NAEGcz+YGciCI2mk;800000fB+dR6$BLo6#@Vi02Krk0ssI2CYC`Q6$%+O1ONsZ7!z-DVrg_?Y(-;ZWo~qGX=7=0c@+#65*Eq=7VZKj>OUr^J{%Yb2&^1hWINA47zlakCNT)Cd_NWdC<q9w99m>M&p#LldFduG2&{ZR02wS51Qi4o0ss{N6$BLm0000c#6=tx3K=y700tQt5N~2^X>oOFba_;7PjX^ubQKI15*Eq=7SIAFm_#P@LL3+f&CJZq%q(U=7zmuiAh2?1p+6P?C<x8W%*@OzW<VGSoWvloa%iDH02wS51Qi4o0ss{N6$BLm0000c&Pf~<3K=y700tQt4{c>(Zd7G$aAk5~bairN6$};<7RUk?vH~WeNG6;|92f`y000000DV9h2mk;8001Ctz!m@~2mk;80001eKo|%B00000AZ)+@87vh96$BLm02Kfg1Qh}R001VmO&k>p88rj|1{oL+b9HiNVPj=ba%FRKb#i4D3>Fd={s9({0w$15CZtOo7zh9W00000S3npD000000001h76>Q^000000037&7zh9W000000Du4)EENP51Qh}R6#x|k6#@VN04AJK92Eo^H3R?#85j+2b!}p0a!+PWZ*OjM6$};<7PJBuk^(05PbP{_92f{s000000000O2mk;80000002Tl!2u}b20000002l}W000000000087vh96$BLm02Kfg1Qh}R000)QG8Pgh2p0A-6aWAj92g8iQc_P>I&))aWo=;?m<%s3FE1}NXGK9lK~X_LK|w)5K}$hFcR@mDK}JDAXF)?}K|w)5K|(@7K}|tHK|w=7K}kVDK|w-6K}kVDK|w-6K|w)6Q9(gLL1#iiK|wS@O+i6RcSJ!!LNq~lK|xbNO+i6IXGB3kL}x)~K|ygrNkKtKQA0sNK{!EYK|w=7NkKtQQ9?mMLt{ZvK|xwUML|JBXF@?iL~}twK|ymtK|w)CXF)+hLP0@6K|w-6K|w)9L^v-mI6*-{MMFhFK|w)7K|w)5K}A79K|w)5K|w)5FhM~<LUm+UF=<*@N>EyBF?M1yHF`loK|xSZdM__`K|w)5MnOSAK|y6udM__`K|w)LO;JHXK|x7DK|w)5Q9(gLK|w)5K|w@!K|w)CPBCL`T4-4^R#9VZR(fMIR(V=cL}f2fFE2qsK|?`HdN?mHI6*-{MMOnGK|w)7K|w)5K}A79K|w)5K|w)5I6*-{LUdYJMs-$4R&--WR(V-iQZZv~RY65|GcPZ9K|w)hL}xEhFE2qsK|(=zK|w)5L2E%lK|w-6K|w)5K|w)5K~X_LK|x|NN^@aYMloSnQZ-gbIZ$J5N^oRBK|w)5LQ5}ZFE2qsK|?`8YB?`2I6*-{MMOnGK|w)7Q9(gLK}A79K|w)5K|w)6O+i6HLUv<qN^oRHMsreEPHI|LH85jsY-(XxQZ+$IOgS$vI6*-{NkLUmGcPZ9K|w)LO;JHXK|xAEK|w)5Q9(gLK|w)5K|w`lK|w)BIeA%aY%p0@cuHbzPIF;bHE3i<Y<6O2K|ytPGcPZ9K|w)hLrE`YFE2qsK|(=zK|w)5L1#fhK|w-6K|w)5K|w)5K~O<KK|x|ON@-zOQfOjET5MuRIci~8QZ+$OH#sjaI6*-{NkMi{GcPZ9K|w)LO;JHXK|x7DK|w)5Q9(gLK|w)5K|w@OK|w)BIeA%aY%p0@cu`tdN^oK_Q9(gLL1`~`FE2qsK|?`ML^&@nI6*-{MMOnGK|w)6cR@ixK}A79K|w)5K|w)5XF)+hLTF=0Xl!CfY<6O2K|w)5LQ^kxFE2qsK|?`9L@O^ZI6*-{MMOnGK|w)6cR@ixK}A79K|w)5K|w)5Z$UvpLU&qNN^oIER&!xjIdwrnLqjiiFE2qsK|?`CL@O^ZI6*-{MMOnGK|w)7K|w)5K}A79K|w)5K|w)5I6*-{LTh1IPB2+lHELQ_T6tnJF;rw%G(mKBH!m-DK|w)hLq#ukFE2qsK|(=zK|w)5L1RHdK|w-6K|w)5K|w)5K}<nGK|x|iR&`o4Qg2i;T18n{Ye6-3H!m-DK|w)hLohFPFE2qsK|(=zK|w)5L1#fhK|w-6K|w)5K|w)5K~F(JK|y0}IWSpRQgC82HBnkvN^oK_Q9(gMaxX70I6*-{NkKMHH!m-DK|w)LO;JHXK|x7DK|w)5Q9(gLK|w)5K|w@kK|w)BR(e@7Mt53PT5MuRIci~8QZ+$ALQgL*FE2qsK|?`8dM__8I6*-{MMOnGK|w)6XF)+hK}A79K|w)5K|w)5V?jYdLT+PiMlxYoQZ+$HaxX70I6*-{NkLOkFE1~5K|w)LO;JHXK|x1BK|w)5Q9(gLK|w)5K|w-yK|w)BQfy*wPH<#MYe8^!FE1~5K|w)hLNG5cFE2qsK|(=zK|w)5L1RHdK|w-6K|w)5K|w)5K}kVDK|xwaY-nOvRxnyvNkKtDK|wKgFE1~5K|w)hLu)TDFE2qsK|(=zK|w)5L1RHdK|w-6K|w)5K|w)5K}A79K|x|yMtWm+K|w-iK}kVDSwTW)K|w)5K|w)hL{%>@FE2qsK|(@OK|w)5L1RHdK|w-6K|w)5K|w)5K}A79K|x||Ms!+HK|w)5K}A79Q9(jMK|w)6K|w}AK}kVDSwTW~K}|tHK|w`9K~X_LK|w)5K|(=jK|w)5L1RHdK|w-6K|w)5K|w)5K|?`7K|x_xQ9(gLXF)_kK}kVDV?jefK|w)5K|w)5MnOSAK|w)5K|w)5Jv{&zA8=uEadl;MEn{$SEn#wUZ+9SeWpZ<AZ*CwqE-)@J8T=F+000~S02}}S8~^|u000~S02}}S8~^|u000~S02}}S8~^|u000~S02}}S8~^|u000~S02}}S8~^|SArJrnP(e~L"
                )
            )
        )


if __name__ == "__main__":  # pragma: no cover

    md = PlanetData().get()

    print(md.schema)
    print(md.shape)
